#!/usr/bin/perl

package Cognition::Export::Contact;
use Cognition::Export;
use Cognition::Export::Address;
use Cognition::Export::Point;
use Data::Dumper;
use HTML::Entities qw(encode_entities_numeric);
use JSON 2.0;
use MIME::Base64;
use URI;

# Most RDF formulations of vCard treat 'N' as a *resource* with predicates
# for each component. Need to support this. (But FOAF does not; internal
# hCard representation does not.)

# Reference:
# http://www.ibiblio.org/hhalpin/homepage/notes/vcardtable.html
# (Though that page does contain a number of mistakes, such as equating a vCard
# "TITLE" [a job title] with a FOAF "title" [e.g. Mr, Mrs].)

sub is
{
	my $subject = shift;
	return unless ($subject);
	return
		unless (defined $subject->property('http://www.w3.org/1999/02/22-rdf-syntax-ns#type'));
	foreach my $Type ( @{$subject
		->property('http://www.w3.org/1999/02/22-rdf-syntax-ns#type')
		->valObjects} )
	{
		my $type = $Type->toString;
		return 1
			if (($type eq 'http://xmlns.com/foaf/0.1/Person')
			||  ($type eq 'http://xmlns.com/foaf/0.1/Organisation')
			||  ($type eq 'http://xmlns.com/foaf/0.1/Project')
			||  ($type eq 'http://xmlns.com/foaf/0.1/Group')
			||  ($type eq 'http://www.w3.org/2001/vcard-rdf/3.0#VCARD')
			||  ($type eq 'http://www.w3.org/2006/vcard/ns#VCard')
			||	 ($type eq 'http://www.w3.org/2006/03/hcard#vcard')
			||  ($type eq 'http://www.w3.org/2000/10/swap/pim/contact#Person')
			||  ($type eq 'http://www.w3.org/2000/10/swap/pim/contact#Male')
			||  ($type eq 'http://www.w3.org/2000/10/swap/pim/contact#Female')
			||  ($type eq 'http://www.w3.org/2000/10/swap/pim/contact#SocialEntity')
			||  ($type eq 'urn:ietf:rfc:2426#Vcard')
			||  ($type eq 'http://xmlns.com/wot/0.1/#term_User'));
	}
	return;
}

sub outputs
{
	return [
		{
			code  => 'VCF',
			name  => 'vCard',
			type  => 'text/directory',
			multi => 1
		},
		{
			code  => 'Name',
			name  => 'Contact Name',
			type  => 'text/plain',
			multi => 1
		},
		{
			code  => 'JSON',
			name  => 'jCard',
			type  => 'text/x-json',
			multi => 1
		}
	];
}

sub multiJSON
{
	my $triples = shift;
	my @rv;
	foreach my $s (keys %$triples)
	{
		if (defined $triples->{$s} && is($triples->{$s}))
		{
			my %r = _outJSON($triples, $s);
			push @rv, \%r;
		}
	}
 	return to_json(\@rv, {utf8 => 1, pretty => 1, canonical => 1})."\n";
}

sub multiVCF
{
	my $triples = shift;
	my $rv = '';
	foreach my $s (keys %$triples)
	{
		$rv .= outVCF($triples, $s)
			if (defined $triples->{$s} && is($triples->{$s}));
	}
	return $rv;
}

sub outJSON
{
	my %json = _outJSON(@_);
 	return to_json(\%json, {utf8 => 1, pretty => 1, canonical => 1})."\n";
}

sub _outJSON
{
	my $triples = shift;
	my $subject = shift;
	my $S = findSubject($triples, $subject);
	my %rv;

	# UID
	my $uid = firstOf($S, 'urn:ietf:rfc:2426#uid',
		'http://www.w3.org/2006/vcard/ns#uid',
		'http://www.w3.org/2006/03/hcard#uid',
		'http://www.w3.org/2001/vcard-rdf/3.0#UID',
		'http://purl.org/dc/terms/identifier',
		'http://purl.org/dc/elements/1.1/identifier');
	$rv{uid} = $uid
		if ($uid);
	
	# Name components
	$rv{n}{'honorific-prefix'} = [allOf($S, 'urn:ietf:rfc:2426#honorific-prefix',
		'http://www.w3.org/2006/vcard/ns#honorific-prefix',
		'http://www.w3.org/2006/03/hcard#honofific-prefix',
		'http://www.w3.org/2001/vcard-rdf/3.0#Prefix',
		'http://xmlns.com/foaf/0.1/title',
		'http://www.w3.org/2000/10/swap/pim/contact#personalTitle')];
	$rv{n}{'given-name'} = [allOf($S, 'urn:ietf:rfc:2426#given-name',
		'http://www.w3.org/2006/vcard/ns#given-name',
		'http://www.w3.org/2006/03/hcard#given-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#Given',
		'http://xmlns.com/foaf/0.1/givenname',
		'http://www.w3.org/2000/10/swap/pim/contact#firstName')];
	$rv{n}{'additional-name'} = [allOf($S, 'urn:ietf:rfc:2426#additional-name',
		'http://www.w3.org/2006/vcard/ns#additional-name',
		'http://www.w3.org/2006/03/hcard#additional-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#Other')];
	$rv{n}{'family-name'} = [allOf($S, 'urn:ietf:rfc:2426#family-name',
		'http://www.w3.org/2006/vcard/ns#family-name',
		'http://www.w3.org/2006/03/hcard#family-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#Family',
		'http://xmlns.com/foaf/0.1/family_name',
		'http://xmlns.com/foaf/0.1/surname',
		'http://www.w3.org/2000/10/swap/pim/contact#lastName')];
	$rv{n}{'honorific-suffix'} = [allOf($S, 'urn:ietf:rfc:2426#honorific-suffix',
		'http://www.w3.org/2006/vcard/ns#honorific-suffix',
		'http://www.w3.org/2006/03/hcard#honorific-suffix',
		'http://www.w3.org/2001/vcard-rdf/3.0#Suffix',
		'http://www.w3.org/2000/10/swap/pim/contact#personalSuffix')];
		
	# Formatted name
	my $fn = firstOf($S, 'urn:ietf:rfc:2426#fn',
		'http://www.w3.org/2006/vcard/ns#fn',
		'http://www.w3.org/2006/03/hcard#fn',
		'http://www.w3.org/2001/vcard-rdf/3.0#FN',
		'http://xmlns.com/foaf/0.1/fullName',
		'http://xmlns.com/foaf/0.1/name',
		'http://www.w3.org/2000/10/swap/pim/contact#knownAs',
		'http://purl.org/dc/terms/title',
		'http://purl.org/dc/elements/1.1/title');	
	if (!$fn)
		{ $fn = $rv{n}{givenName}[0].' '.$rv{n}{familyName}[0]; }
	$rv{fn} = "$fn";

	# Nickname
	$rv{nickname} = [allOf($S, 'urn:ietf:rfc:2426#nickname',
		'http://www.w3.org/2006/vcard/ns#nickname',
		'http://www.w3.org/2006/03/hcard#nickname',
		'http://www.w3.org/2001/vcard-rdf/3.0#NICKNAME',
		'http://xmlns.com/foaf/0.1/nick',
		'http://www.w3.org/2000/10/swap/pim/contact#knownAs')];

	# Sort-String
	my $sortstring = firstOf($S, 'urn:ietf:rfc:2426#sort-string',
		'http://www.w3.org/2006/vcard/ns#sort-string',
		'http://www.w3.org/2006/03/hcard#sort-string',
		'http://www.w3.org/2001/vcard-rdf/3.0#SORT-STRING',
		'http://www.w3.org/2000/10/swap/pim/contact#sortName');
	$rv{'sort-string'} = $sortstring
		if ($sortstring);

	# Title
	$rv{title} = [allOf($S, 'urn:ietf:rfc:2426#title',
		'http://www.w3.org/2006/vcard/ns#title',
		'http://www.w3.org/2006/03/hcard#title',
		'http://www.w3.org/2001/vcard-rdf/3.0#TITLE')];

	# Role
	$rv{role} = [allOf($S, 'urn:ietf:rfc:2426#role',
		'http://www.w3.org/2006/vcard/ns#role',
		'http://www.w3.org/2006/03/hcard#role',
		'http://www.w3.org/2001/vcard-rdf/3.0#ROLE')];

	# Addresses
	# TODO: Imply "type" from homeAdr, workAdr.
	my @adrs = allOfObj($S, 'urn:ietf:rfc:2426#adr',
		'http://www.w3.org/2006/vcard/ns#adr',
		'http://www.w3.org/2006/vcard/ns#homeAdr',
		'http://www.w3.org/2006/vcard/ns#workAdr',
		'http://www.w3.org/2006/vcard/ns#unlabelledAdr',
		'http://www.w3.org/2006/03/hcard#adr',
		'http://www.w3.org/2001/vcard-rdf/3.0#ADR',
		'http://www.w3.org/2000/10/swap/pim/contact#address');
	foreach my $adr (@adrs)
	{
		my $adrSubject = findSubject($triples, $adr->toString);
		if ($adrSubject)
		{
			if (Cognition::Export::Address::is($adrSubject))
			{
				%a = Cognition::Export::Address::_outJSON($triples, $adrSubject->uri);
				push @{$rv{adr}}, \%a;
			}
		}
		else
		{
			my $a = $adr->toString;
			$a =~ s/(^\s+|\s+$)//g;
			$a =~ s/\s+/ /g;
			push @{$rv{label}}, $a if ($a);
		}
	}

	# Geo
	my $geo = findSubject($triples, firstOf($S, 'urn:ietf:rfc:2426#geo',
		'http://www.w3.org/2006/vcard/ns#geo',
		'http://www.w3.org/2006/03/hcard#geo',
		'http://www.w3.org/2001/vcard-rdf/3.0#GEO'));
	if ($geo && Cognition::Export::Point::is($geo))
	{
		my %g = Cognition::Export::Point::_outJSON($triples, $geo->uri);
		$rv{'geo'} = \%g;
	}

	# Labels
	push @{$rv{label}}, allOf($S, 'urn:ietf:rfc:2426#label',
		'http://www.w3.org/2006/vcard/ns#label',
		'http://www.w3.org/2006/03/hcard#label',
		'http://www.w3.org/2001/vcard-rdf/3.0#LABEL');

	# Telephone numbers
	my @tels = allOfObj($S, 'urn:ietf:rfc:2426#tel',
		'http://www.w3.org/2006/vcard/ns#tel',
		'http://www.w3.org/2006/vcard/ns#homeTel',
		'http://www.w3.org/2006/vcard/ns#workTel',
		'http://www.w3.org/2006/vcard/ns#mobileTel',
		'http://www.w3.org/2006/vcard/ns#unlabelledTel',
		'http://www.w3.org/2006/vcard/ns#fax',
		'http://www.w3.org/2006/03/hcard#tel',
		'http://www.w3.org/2001/vcard-rdf/3.0#TEL',
		'http://www.w3.org/2000/10/swap/pim/contact#phone',
		'http://www.w3.org/2000/10/swap/pim/contact#mobile',
		'http://xmlns.com/foaf/0.1/phone');
	foreach my $tel (@tels)
	{
		my $telSubject = findSubject($triples, $tel->toString);

		my $usages;
		$usages = 'CELL'
			if ($tel->property->uri eq 'http://www.w3.org/2006/vcard/ns#mobileTel');
		$usages = 'CELL'
			if ($tel->property->uri eq 'http://www.w3.org/2000/10/swap/pim/contact#mobile');
		$usages = 'FAX'
			if ($tel->property->uri eq 'http://www.w3.org/2006/vcard/ns#fax');
		$usages = 'HOME'
			if ($tel->property->uri eq 'http://www.w3.org/2006/vcard/ns#homeTel');
		$usages = 'WORK'
			if ($tel->property->uri eq 'http://www.w3.org/2006/vcard/ns#workTel');
			
		if ($telSubject)
		{
			my @usages = allOf($telSubject, 'urn:ietf:rfc:2426#type');
			$usages = uc(join ',', @usages);
		}
		
		$usages = lc($usages);
		my @usages = split /\,/, $usages;
		
		my $telno = $tel->toString;
		if ($telno =~ /^tel\:/)
		{
			$telno =~ s/^tel\://;
			$telno =~ s/\;.*$//g;
		}
		else
		{
			# TODO: allow <rdf:value> here.
			$telno = firstOf($telSubject, 'urn:ietf:rfc:2426#value');
		}
		
		if (!$usages) { $usages = 'voice'; }

		my %tel;
		$tel{'value'} = $telno;
		$tel{'type'} = [@usages];
		push @{$rv{'tel'}}, \%tel;
	}

	# E-mail addresses. W3C vocab defines a bunch of weird e-mail types which
	# have nothing to do with vCard. This procedure is almost identical to "TEL:"
	# but I don't think we can wrap it all in a function.
	my @emails = allOfObj($S, 'urn:ietf:rfc:2426#email',
		'http://www.w3.org/2006/vcard/ns#email',
		'http://www.w3.org/2006/vcard/ns#workEmail',
		'http://www.w3.org/2006/vcard/ns#mobileEmail',
		'http://www.w3.org/2006/vcard/ns#unlabelledEmail',
		'http://www.w3.org/2006/03/hcard#email',
		'http://www.w3.org/2001/vcard-rdf/3.0#EMAIL',
		'http://www.w3.org/2000/10/swap/pim/contact#emailAddress',
		'http://xmlns.com/foaf/0.1/mbox',
		'http://xmlns.com/foaf/0.1/mbox_sha1sum');
	foreach my $email (@emails)
	{
		my $emailSubject = findSubject($triples, $email->toString);

		my $usages;
		if ($emailSubject)
		{
			my @usages = allOf($emailSubject, 'urn:ietf:rfc:2426#type');
			$usages = uc(join ',', @usages);
		}
		
		my $emailno = $email->toString;
		$usages = 'X-SHA1SUM'
			if ($email->property->uri eq 'http://xmlns.com/foaf/0.1/mbox_sha1sum');
		if ($emailno =~ /^mailto\:/)
		{
			$emailno =~ s/^mailto\://;
			$emailno =~ s/\?.*$//g;
		}
		else
		{
			# TODO: allow <rdf:value> here.
			$emailno = firstOf($emailSubject, 'urn:ietf:rfc:2426#value');
		}
		
		if (!$usages) { $usages = 'internet'; }
		
		my @usages = split /\,/, lc($usages);
		my %tel;
		$tel{'value'} = $emailno;
		$tel{'type'} = [@usages];
		push @{$rv{'email'}}, \%tel;
	}

	# URLs - Note: W3C's 2001 vCard RDF vocab does not contain URL!
	my @urls = allOf($S, 'urn:ietf:rfc:2426#url',
		'http://www.w3.org/2006/vcard/ns#url',
		'http://www.w3.org/2006/03/hcard#url',
		'http://xmlns.com/foaf/0.1/homepage',
		'http://xmlns.com/foaf/0.1/weblog',
		'http://xmlns.com/foaf/0.1/workplaceHomepage',
		'http://xmlns.com/foaf/0.1/workInfoHomepage',
		'http://xmlns.com/foaf/0.1/schoolHomepage',
		'http://xmlns.com/foaf/0.1/tipjar',
		'http://www.w3.org/2000/10/swap/pim/contact#homepage',
		'http://www.w3.org/2000/10/swap/pim/contact#publicHomepage');
	foreach my $url (@urls)
	{
		push @{$rv{'url'}}, $url;
		if ($url =~ /^(sip|xmpp|irc|ymsgr|msn|aim|im|pres|msnim|skype)\:/i)
			{ push @{$rv{'impp'}}, { value => $url }; }
	}

	# IMPP
	my @impps = allOfObj($S, 'urn:ietf:rfc:4770#impp');
	foreach my $impp (@impps)
	{
		my ($usages, @usages);
		my $imppSubject = findSubject($triples, $impp->toString);
		
		if ($imppSubject)
		{
			@usages = allOf($imppSubject, 'urn:ietf:rfc:4770#type');
			$usages = uc(join ',', @usages);
			@usages = split /\,/, lc($usages);
		}
		
		my $i = firstOf($imppSubject, 'urn:ietf:rfc:4770#value');
		my %impp;
		if (length $i)
		{
			$impp{'value'} = "$i";
			$impp{'type'} = [@usages];
			push @{$rv{'impp'}}, \%impp;
		}
	}
	
	my %imppMap = (
		'http://xmlns.com/foaf/0.1/aimChatID'    => 'aim:goim?screenname=%s',
		'http://xmlns.com/foaf/0.1/icqChatID'    => 'http://www.icq.com/people/cmd.php?uin=%d&action=message',
		'http://xmlns.com/foaf/0.1/jabberID'     => 'xmpp:%s',
		'http://xmlns.com/foaf/0.1/msnChatID'    => 'msnim:chat?contact=%s',
		'http://xmlns.com/foaf/0.1/yahooChatID'  => 'ymsgr:sendIM?%s'
	);
	foreach my $pred (keys %imppMap)
	{
		my @impps = allOf($S, $pred);
		foreach my $impp (@impps)
		{
			push @{$rv{'impp'}}, { value => sprintf($imppMap{$pred}, $impp) };
		}
	}
	
	# Org
	my @orgs = allOfObj($S, 'urn:ietf:rfc:2426#org',
		'http://www.w3.org/2006/vcard/ns#org',
		'http://www.w3.org/2006/03/hcard#org',
		'http://www.w3.org/2001/vcard-rdf/3.0#ORG');
	foreach my $org (@orgs)
	{
		my $orgSubject = findSubject($triples, $org->toString);
		
		if ($orgSubject)
		{
			my %o;
		
			my $orgstring = firstOf($orgSubject,
				'urn:ietf:rfc:2426#organization-name',
				'http://www.w3.org/2006/vcard/ns#organization-name',
				'http://www.w3.org/2006/03/hcard#organization-name',
				'http://www.w3.org/2001/vcard-rdf/3.0#Orgname');
			$o{'organization-name'} = "$orgstring";

			my @units = allOf($orgSubject,
				'urn:ietf:rfc:2426#organization-unit',
				'http://www.w3.org/2006/vcard/ns#organization-unit',
				'http://www.w3.org/2006/03/hcard#organization-unit',
				'http://www.w3.org/2001/vcard-rdf/3.0#Orgunit');
			$o{'organization-unit'} = [@units] if (@units);

			push @{$rv{'org'}}, \%o;
		}
		else
		{
			push @{$rv{'org'}}, $org->toString;
		}
	}

	# Mailer
	$rv{'mailer'} = [allOf($S, 'urn:ietf:rfc:2426#mailer',
		'http://www.w3.org/2006/vcard/ns#mailer',
		'http://www.w3.org/2006/03/hcard#mailer',
		'http://www.w3.org/2001/vcard-rdf/3.0#MAILER')];

	# Photo
	$rv{'photo'} = [allOf($S, 'urn:ietf:rfc:2426#photo',
		'http://www.w3.org/2006/vcard/ns#photo',
		'http://www.w3.org/2006/03/hcard#photo',
		'http://www.w3.org/2001/vcard-rdf/3.0#PHOTO',
		'http://xmlns.com/foaf/0.1/img',
		'http://xmlns.com/foaf/0.1/depiction')];

	# Logo
	$rv{'logo'} = [allOf($S, 'urn:ietf:rfc:2426#logo',
		'http://www.w3.org/2006/vcard/ns#logo',
		'http://www.w3.org/2006/03/hcard#logo',
		'http://www.w3.org/2001/vcard-rdf/3.0#LOGO',
		'http://xmlns.com/foaf/0.1/logo')];

	# Sound
	$rv{'sound'} = [allOf($S, 'urn:ietf:rfc:2426#sound',
		'http://www.w3.org/2006/vcard/ns#sound',
		'http://www.w3.org/2006/03/hcard#sound',
		'http://www.w3.org/2001/vcard-rdf/3.0#SOUND')];

	# Key (crypto)
	$rv{'key'} = [allOf($S, 'urn:ietf:rfc:2426#key',
		'http://www.w3.org/2006/vcard/ns#key',
		'http://www.w3.org/2006/03/hcard#key',
		'http://www.w3.org/2001/vcard-rdf/3.0#KEY')];
	
	# Keys (WOT RDF schema)
	my @keys = allOfObj($S, 'http://xmlns.com/wot/0.1/hasKey');
	foreach my $key (@keys)
	{
		my $keySubject = findSubject($triples, $key->toString);
		if ($keySubject)
		{
			my $uri = firstOf($keySubject, 'http://xmlns.com/wot/0.1/pubkeyAddress');
			push @{$rv{'key'}}, $uri
				if ($uri);
		}
	}

	# Notes
	$rv{'note'} = [allOf($S, 'urn:ietf:rfc:2426#note',
		'http://www.w3.org/2006/vcard/ns#note',
		'http://www.w3.org/2006/03/hcard#note',
		'http://www.w3.org/2001/vcard-rdf/3.0#NOTE',
		'http://www.w3.org/2000/01/rdf-schema#comment',
		'http://purl.org/dc/terms/abstract',
		'http://purl.org/dc/terms/description',
		'http://purl.org/dc/elements/1.1/description',
		'http://xmlns.com/foaf/0.1/plan',
		'http://xmlns.com/foaf/0.1/geekcode')];

 	# Agents
 	my @agents = allOfObj($S, 'urn:ietf:rfc:2426#agent',
 		'http://www.w3.org/2001/vcard-rdf/3.0#AGENT',
 		'http://www.w3.org/2006/vcard/ns#agent',
 		'http://www.w3.org/2006/03/hcard#agent');
 	foreach my $agent (@agents)
	{
		my $agentSubject = findSubject($triples, $agent->toString);
		if ($agentSubject)
		{
			if (is($agentSubject))
			{
				my %a = _outJSON($triples, $agentSubject->uri);
				push @{$rv{'agent'}}, \%a;
			}
			else
			{
				my $a = firstOf($agentSubject,
					'urn:ietf:rfc:2426#fn',
					'http://www.w3.org/2006/vcard/ns#fn',
					'http://www.w3.org/2006/03/hcard#fn',
					'http://www.w3.org/2001/vcard-rdf/3.0#FN',
					'http://xmlns.com/foaf/0.1/fullName',
					'http://xmlns.com/foaf/0.1/name',
					'http://www.w3.org/2000/10/swap/pim/contact#knownAs',
					'http://purl.org/dc/terms/title',
					'http://purl.org/dc/elements/1.1/title');
				push @{$rv{'agent'}}, $a if ($a);
			}
		}
		else
		{
			my $a = $agent->toString;
			$a =~ s/(^\s+|\s+$)//g;
			$a =~ s/\s+/ /g;
			push @{$rv{'agent'}}, $a if ($a);
		}
	}

	# TZ (timezone)
	my $tz = firstOf($S, 'urn:ietf:rfc:2426#tz',
		'http://www.w3.org/2006/vcard/ns#tz',
		'http://www.w3.org/2006/03/hcard#tz',
		'http://www.w3.org/2001/vcard-rdf/3.0#TZ');
	$rv{'tz'} = $tz if ($tz);

	# Bday (date of birth)
	my $bday = firstOf($S, 'urn:ietf:rfc:2426#bday',
		'http://www.w3.org/2006/vcard/ns#bday',
		'http://www.w3.org/2006/03/hcard#bday',
		'http://www.w3.org/2001/vcard-rdf/3.0#BDAY',
		'http://www.w3.org/2000/10/swap/pim/contact#birthday',
		'http://xmlns.com/foaf/0.1/birthday');
	$rv{'bday'} = $bday if ($bday);

	# Dday (date of birth)
	my $dday = firstOf($S, 'urn:ietf:rfc:2426#dday');
	$rv{'dday'} = $dday if ($dday);
		
	# Gender
	my $gender = firstOf($S, 'urn:ietf:rfc:2426#gender',
		'http://xmlns.com/foaf/0.1/gender');
	if ($gender)
	{
		$rv{'gender'} = $gender;
	}
	else
	{
		if (inProp($S, 'http://www.w3.org/2000/10/swap/pim/contact#Male',
			'http://www.w3.org/1999/02/22-rdf-syntax-ns#type'))
		{
			$rv{'gender'} = 'Male';
		}

		if (inProp($S, 'http://www.w3.org/2000/10/swap/pim/contact#Female',
			'http://www.w3.org/1999/02/22-rdf-syntax-ns#type'))
		{
			$rv{'gender'} = 'Female';
		}
	}
	
	# Categories
	my @cats = allOfObj($S, 'urn:ietf:rfc:2426#category',
		'http://www.w3.org/2006/vcard/ns#category',
		'http://www.w3.org/2006/03/hcard#category',
		'http://www.w3.org/2001/vcard-rdf/3.0#CATEGORIES');
	my $catstring = '';
	foreach my $cat (@cats)
	{
		my $catSubject = findSubject($triples, $cat->toString);
		if ($catSubject)
		{
			my $category = firstOf($catSubject,
				'http://www.w3.org/1999/02/22-rdf-syntax-ns#value',
				'http://purl.org/dc/terms/description',
				'http://purl.org/dc/elements/1.1/description');
			$catstring .= ','.uc($category)
				if (length $category);
		}
		else
		{
			my $category = $cat->toString;
			$category =~ s/(^\s+|\s+$)//g;
			$category =~ s/\s+/ /g;
			$catstring .= ','.uc($category);
		}
	}
	$catstring =~ s/^\,//;
	$rv{'categories'} = [split /\,/, $catstring]
		if ($catstring);
	$rv{'category'} = [split /\,/, $catstring]
		if ($catstring);

	# Class
	my $class = uc(firstOf($S, 'urn:ietf:rfc:2426#class',
		'http://www.w3.org/2006/vcard/ns#class',
		'http://www.w3.org/2006/03/hcard#class',
		'http://www.w3.org/2001/vcard-rdf/3.0#CLASS'));
	$class =~ s/\s//g;
	$rv{'class'} = lc($class)
		if ($class =~ /^(PUBLIC|PRIVATE|CONFIDENTIAL|X-)/);
		
	# Rev
	my $rev = uc(firstOf($S, 'urn:ietf:rfc:2426#rev',
		'http://www.w3.org/2006/vcard/ns#rev',
		'http://www.w3.org/2006/03/hcard#rev',
		'http://www.w3.org/2001/vcard-rdf/3.0#REV'));
	$rv{'rev'} = $rev
		if ($rev);

	# Birth - accept vCard/adr/geo?
	my $birth = firstOf($S, 'urn:ietf:rfc:2426#birth');
	$rv{'birth'} = $birth
		if ($birth);

	# Death - accept vCard/adr/geo?
	my $death = firstOf($S, 'urn:ietf:rfc:2426#death');
	$rv{'death'} = $death
		if ($death);

	# Lang
	# TODO: there are a bunch of schemas out there for speaks, writes & reads.
	my $lang = firstOf($S, 'urn:ietf:rfc:2426#lang');
	$rv{'lang'} = $lang
		if ($lang);

	# Kind
	my $kind = firstOf($S, 'urn:ietf:rfc:2426#kind');
	$rv{'kind'} = $kind
		if ($kind);

	# Remove empty leafnodes from structure.	
	foreach $k (keys %rv)
	{
		if (!defined $rv{$k})
		{
			delete $rv{$k};
		}
		elsif ((ref $rv{$k}) eq 'ARRAY')
		{
			delete $rv{$k} unless scalar(@{$rv{$k}});
		}
	}

	# Remove empty leafnodes from name parts.	
	foreach $k (keys %{$rv{'n'}})
	{
		delete $rv{'n'}{$k} unless scalar(@{$rv{'n'}{$k}});
	}
	
	return %rv;
}

sub outVCF
{
	my $triples = shift;
	my $subject = shift;
	my $S = findSubject($triples, $subject);
	my @rv = (['BEGIN','VCARD']);

	# UID
	my $uid = firstOf($S, 'urn:ietf:rfc:2426#uid',
		'http://www.w3.org/2006/vcard/ns#uid',
		'http://www.w3.org/2006/03/hcard#uid',
		'http://www.w3.org/2001/vcard-rdf/3.0#UID',
		'http://purl.org/dc/terms/identifier',
		'http://purl.org/dc/elements/1.1/identifier');
	push @rv, ['UID', $uid]
		if ($uid);
	
	# Name components
	my $honorific_pfx = firstOf($S, 'urn:ietf:rfc:2426#honorific-prefix',
		'http://www.w3.org/2006/vcard/ns#honorific-prefix',
		'http://www.w3.org/2006/03/hcard#honofific-prefix',
		'http://www.w3.org/2001/vcard-rdf/3.0#Prefix',
		'http://xmlns.com/foaf/0.1/title',
		'http://www.w3.org/2000/10/swap/pim/contact#personalTitle');
	my $firstname = firstOf($S, 'urn:ietf:rfc:2426#given-name',
		'http://www.w3.org/2006/vcard/ns#given-name',
		'http://www.w3.org/2006/03/hcard#given-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#Given',
		'http://xmlns.com/foaf/0.1/givenname',
		'http://www.w3.org/2000/10/swap/pim/contact#firstName');
	my $middlename = firstOf($S, 'urn:ietf:rfc:2426#additional-name',
		'http://www.w3.org/2006/vcard/ns#additional-name',
		'http://www.w3.org/2006/03/hcard#additional-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#Other');
	my $surname = firstOf($S, 'urn:ietf:rfc:2426#family-name',
		'http://www.w3.org/2006/vcard/ns#family-name',
		'http://www.w3.org/2006/03/hcard#family-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#Family',
		'http://xmlns.com/foaf/0.1/family_name',
		'http://xmlns.com/foaf/0.1/surname',
		'http://www.w3.org/2000/10/swap/pim/contact#lastName');
	my $honorific_sfx = firstOf($S, 'urn:ietf:rfc:2426#honorific-suffix',
		'http://www.w3.org/2006/vcard/ns#honorific-suffix',
		'http://www.w3.org/2006/03/hcard#honorific-suffix',
		'http://www.w3.org/2001/vcard-rdf/3.0#Suffix',
		'http://www.w3.org/2000/10/swap/pim/contact#personalSuffix');
		
	# Formatted name
	my $fn = firstOf($S, 'urn:ietf:rfc:2426#fn',
		'http://www.w3.org/2006/vcard/ns#fn',
		'http://www.w3.org/2006/03/hcard#fn',
		'http://www.w3.org/2001/vcard-rdf/3.0#FN',
		'http://xmlns.com/foaf/0.1/fullName',
		'http://xmlns.com/foaf/0.1/name',
		'http://www.w3.org/2000/10/swap/pim/contact#knownAs',
		'http://purl.org/dc/terms/title',
		'http://purl.org/dc/elements/1.1/title');	
	if (!$fn)
		{ $fn = "$firstname $surname"; }
	push @rv, ['FN',$fn];

	# Structured name
	my $n = firstOf($S, 'urn:ietf:rfc:2426#n',
		'http://www.w3.org/2006/vcard/ns#n',
		'http://www.w3.org/2006/03/hcard#n',
		'http://www.w3.org/2001/vcard-rdf/3.0#NAME');	
	if (!$n)
	{
		$n = "$surname;$firstname;$middlename;$honorific_pfx;$honorific_sfx";
		$n =~ s/ /\,/;
	}
	push @rv, ['N',$n,1];
	
	# Nickname
	my @nicknames = allOf($S, 'urn:ietf:rfc:2426#nickname',
		'http://www.w3.org/2006/vcard/ns#nickname',
		'http://www.w3.org/2006/03/hcard#nickname',
		'http://www.w3.org/2001/vcard-rdf/3.0#NICKNAME',
		'http://xmlns.com/foaf/0.1/nick',
		'http://www.w3.org/2000/10/swap/pim/contact#knownAs');
	foreach my $nickname (@nicknames)
		{ push @rv, ['NICKNAME',$nickname]; }

	# Sort-String
	my $sortstring = firstOf($S, 'urn:ietf:rfc:2426#sort-string',
		'http://www.w3.org/2006/vcard/ns#sort-string',
		'http://www.w3.org/2006/03/hcard#SORT-STRING',
		'http://www.w3.org/2001/vcard-rdf/3.0#sort-string',
		'http://www.w3.org/2000/10/swap/pim/contact#sortName');
	push @rv, ['SORT-STRING',$sortstring]
		if ($sortstring);

	# Title
	my @titles = allOf($S, 'urn:ietf:rfc:2426#title',
		'http://www.w3.org/2006/vcard/ns#title',
		'http://www.w3.org/2006/03/hcard#title',
		'http://www.w3.org/2001/vcard-rdf/3.0#TITLE');
	foreach my $title (@titles)
		{ push @rv, ['TITLE',$title]; }

	# Role
	my @roles = allOf($S, 'urn:ietf:rfc:2426#role',
		'http://www.w3.org/2006/vcard/ns#role',
		'http://www.w3.org/2006/03/hcard#role',
		'http://www.w3.org/2001/vcard-rdf/3.0#ROLE');
	foreach my $role (@roles)
		{ push @rv, ['ROLE',$role]; }

	# Addresses
	# TODO: homeAdr, workAdr imply types.
	my @adrs = allOfObj($S, 'urn:ietf:rfc:2426#adr',
		'http://www.w3.org/2006/vcard/ns#adr',
		'http://www.w3.org/2006/vcard/ns#homeAdr',
		'http://www.w3.org/2006/vcard/ns#workAdr',
		'http://www.w3.org/2006/vcard/ns#unlabelledAdr',
		'http://www.w3.org/2006/03/hcard#adr',
		'http://www.w3.org/2001/vcard-rdf/3.0#ADR',
		'http://www.w3.org/2000/10/swap/pim/contact#address');
	foreach my $adr (@adrs)
	{
		my $adrSubject = findSubject($triples, $adr->toString);
		if ($adrSubject)
		{
			if (Cognition::Export::Address::is($adrSubject))
			{
				my ($key, $val) = Cognition::Export::Address::outVCFLine($triples, $adrSubject->uri);
				push @rv, [$key, $val, 1];
			}
		}
		elsif ($adr->type)
		{
			push @rv, ['LABEL;VALUE=uri',$adr->toString];
		}
		else
		{
			my $a = $adr->toString;
			$a =~ s/(^\s+|\s+$)//g;
			$a =~ s/\s+/ /g;
			push @rv, ['LABEL',$a] if ($a);
		}
	}

	# Labels
	my @labels = allOf($S, 'urn:ietf:rfc:2426#label',
		'http://www.w3.org/2006/vcard/ns#label',
		'http://www.w3.org/2006/03/hcard#label',
		'http://www.w3.org/2001/vcard-rdf/3.0#LABEL');
	foreach my $label (@labels)
		{ push @rv, ['LABEL',$label]; }

	# Geo
	my @geos = allOfObj($S, 'urn:ietf:rfc:2426#geo',
		'http://www.w3.org/2006/vcard/ns#geo',
		'http://www.w3.org/2006/03/hcard#geo',
		'http://www.w3.org/2001/vcard-rdf/3.0#GEO');
	foreach my $geo (@geos)
	{
		my $geoSubject = findSubject($triples, $geo->toString);
		if ($geoSubject)
		{
			if (Cognition::Export::Point::is($geoSubject))
			{
				my ($key, $val) = Cognition::Export::Point::outVCFLine($triples, $geoSubject->uri);
				push @rv, [$key, $val, 1];
			}
		}
		# TODO: else (support plain string)
	}

	# Telephone numbers
	my @tels = allOfObj($S, 'urn:ietf:rfc:2426#tel',
		'http://www.w3.org/2006/vcard/ns#tel',
		'http://www.w3.org/2006/vcard/ns#homeTel',
		'http://www.w3.org/2006/vcard/ns#workTel',
		'http://www.w3.org/2006/vcard/ns#mobileTel',
		'http://www.w3.org/2006/vcard/ns#unlabelledTel',
		'http://www.w3.org/2006/vcard/ns#fax',
		'http://www.w3.org/2006/03/hcard#tel',
		'http://www.w3.org/2001/vcard-rdf/3.0#TEL',
		'http://www.w3.org/2000/10/swap/pim/contact#phone',
		'http://www.w3.org/2000/10/swap/pim/contact#mobile',
		'http://xmlns.com/foaf/0.1/phone');
	foreach my $tel (@tels)
	{
		my $telSubject = findSubject($triples, $tel->toString);

		my $usages;
		$usages = 'CELL'
			if ($tel->property->uri eq 'http://www.w3.org/2006/vcard/ns#mobileTel');
		$usages = 'CELL'
			if ($tel->property->uri eq 'http://www.w3.org/2000/10/swap/pim/contact#mobile');
		$usages = 'FAX'
			if ($tel->property->uri eq 'http://www.w3.org/2006/vcard/ns#fax');
		$usages = 'HOME'
			if ($tel->property->uri eq 'http://www.w3.org/2006/vcard/ns#homeTel');
		$usages = 'WORK'
			if ($tel->property->uri eq 'http://www.w3.org/2006/vcard/ns#workTel');
			
		if ($telSubject)
		{
			my @usages = allOf($telSubject, 'urn:ietf:rfc:2426#type');
			$usages = uc(join ',', @usages);
		}
		
		my $telno = $tel->toString;
		if ($telno =~ /^tel\:/)
		{
			$telno =~ s/^tel\://;
			$telno =~ s/\;.*$//g;
		}
		else
		{
			# TODO: rdf:value
			$telno = firstOf($telSubject, 'urn:ietf:rfc:2426#value');
		}
		
		if ($usages)
			{ push @rv, ["TEL;TYPE=$usages",$telno]; }
		else
			{ push @rv, ['TEL', $telno]; }
	}

	# E-mail addresses. W3C vocab defines a bunch of weird e-mail types which
	# have nothing to do with vCard. This procedure is almost identical to "TEL:"
	# but I don't think we can wrap it all in a function.
	my @emails = allOfObj($S, 'urn:ietf:rfc:2426#email',
		'http://www.w3.org/2006/vcard/ns#email',
		'http://www.w3.org/2006/vcard/ns#workEmail',
		'http://www.w3.org/2006/vcard/ns#mobileEmail',
		'http://www.w3.org/2006/vcard/ns#unlabelledEmail',
		'http://www.w3.org/2006/03/hcard#email',
		'http://www.w3.org/2001/vcard-rdf/3.0#EMAIL',
		'http://www.w3.org/2000/10/swap/pim/contact#emailAddress',
		'http://xmlns.com/foaf/0.1/mbox',
		'http://xmlns.com/foaf/0.1/mbox_sha1sum');
	foreach my $email (@emails)
	{
		my $emailSubject = findSubject($triples, $email->toString);

		my $usages;
		if ($emailSubject)
		{
			my @usages = allOf($emailSubject, 'urn:ietf:rfc:2426#type');
			$usages = uc(join ',', @usages);
		}
		
		my $emailno = $email->toString;
		$usages = 'X-SHA1SUM'
			if ($email->property->uri eq 'http://xmlns.com/foaf/0.1/mbox_sha1sum');
		if ($emailno =~ /^mailto\:/)
		{
			$emailno =~ s/^mailto\://;
			$emailno =~ s/\?.*$//g;
		}
		else
		{
			# TODO: rdf:value
			$emailno = firstOf($emailSubject, 'urn:ietf:rfc:2426#value');
		}
		
		if ($usages)
			{ push @rv, ["EMAIL;TYPE=$usages",$emailno]; }
		else
			{ push @rv, ['EMAIL',$emailno]; }
	}

	# URLs
	my @urls = allOf($S, 'urn:ietf:rfc:2426#url',
		'http://www.w3.org/2006/vcard/ns#url',
		'http://www.w3.org/2006/03/hcard#url',
		'http://xmlns.com/foaf/0.1/homepage',
		'http://xmlns.com/foaf/0.1/weblog',
		'http://xmlns.com/foaf/0.1/workplaceHomepage',
		'http://xmlns.com/foaf/0.1/workInfoHomepage',
		'http://xmlns.com/foaf/0.1/schoolHomepage',
		'http://xmlns.com/foaf/0.1/tipjar',
		'http://www.w3.org/2000/10/swap/pim/contact#homepage',
		'http://www.w3.org/2000/10/swap/pim/contact#publicHomepage');
	foreach my $url (@urls)
	{
		push @rv, ['URL', $url];
		push @rv, ['IMPP', $url]
			if ($url =~ /^(sip|xmpp|irc|ymsgr|msn|aim|im|pres|msnim|skype)\:/i);
	}

	# IMPP
	my @impps = allOfObj($S, 'urn:ietf:rfc:4770#impp');
	foreach my $impp (@impps)
	{
		my $usages;
		my $imppSubject = findSubject($triples, $impp);
		if ($imppSubject)
		{
			my @usages = allOf($imppSubject, 'urn:ietf:rfc:4770#type');
			$usages = uc(join ',', @usages);
		}
		
		my $i = firstOf($imppSubject, 'urn:ietf:rfc:4770#value');
		if (length $i)
		{
			if (length $usages)
				{ push @rv, ['IMPP;TYPE='.$usages, $i]; }
			else
				{ push @rv, ['IMPP', $i]; }
		}
	}
	
	my %imppMap = (
		'http://xmlns.com/foaf/0.1/aimChatID'   => 'aim:goim?screenname=%s',
		'http://xmlns.com/foaf/0.1/icqChatID'   => 'http://www.icq.com/people/cmd.php?uin=%d&action=message',
		'http://xmlns.com/foaf/0.1/jabberID'    => 'xmpp:%s',
		'http://xmlns.com/foaf/0.1/msnChatID'   => 'msnim:chat?contact=%s',
		'http://xmlns.com/foaf/0.1/yahooChatID' => 'ymsgr:sendIM?%s'
	);
	foreach my $pred (keys %imppMap)
	{
		my @impps = allOf($S, $pred);
		foreach my $impp (@impps)
		{
			push @rv, ['IMPP', sprintf($imppMap{$pred}, $impp)];
		}
	}

	# Org
	my @orgs = allOfObj($S, 'urn:ietf:rfc:2426#org',
		'http://www.w3.org/2006/vcard/ns#org',
		'http://www.w3.org/2006/03/hcard#org',
		'http://www.w3.org/2001/vcard-rdf/3.0#ORG');
	foreach my $org (@orgs)
	{
		my ($orgSubject, $orgstring);
		$orgSubject = findSubject($triples, $org->toString);
		
		if ($orgSubject)
		{
			$orgstring = firstOf($orgSubject,
				'urn:ietf:rfc:2426#organization-name',
				'http://www.w3.org/2006/vcard/ns#organization-name',
				'http://www.w3.org/2006/03/hcard#organization-name',
				'http://www.w3.org/2001/vcard-rdf/3.0#Orgname');
			$orgstring =~ s/\;/\\\;/g;
			$orgstring =~ s/\,/\\\,/g;

			my @units = allOf($orgSubject,
				'urn:ietf:rfc:2426#organization-unit',
				'http://www.w3.org/2006/vcard/ns#organization-unit',
				'http://www.w3.org/2006/03/hcard#organization-unit',
				'http://www.w3.org/2001/vcard-rdf/3.0#Orgunit');
			foreach my $ou (@units)
			{
				$ou =~ s/\;/\\\;/g;
				$ou =~ s/\,/\\\,/g;
				$orgstring .= "; $ou";
			}
		}
		
		$orgstring = $org->toString
			unless (length $orgstring);

		if ($orgstring)
		{
			push @rv, ['ORG', $orgstring, ($orgSubject?1:0)];
		}
	}

	# Mailer
	my @mailers = allOf($S, 'urn:ietf:rfc:2426#mailer',
		'http://www.w3.org/2006/vcard/ns#mailer',
		'http://www.w3.org/2006/03/hcard#mailer',
		'http://www.w3.org/2001/vcard-rdf/3.0#MAILER');
	foreach my $mailer (@mailers)
		{ push @rv, ['MAILER',$mailer]; }

	# Photo
	my @photos = allOf($S, 'urn:ietf:rfc:2426#photo',
		'http://www.w3.org/2006/vcard/ns#photo',
		'http://www.w3.org/2006/03/hcard#photo',
		'http://www.w3.org/2001/vcard-rdf/3.0#PHOTO',
		'http://xmlns.com/foaf/0.1/img',
		'http://xmlns.com/foaf/0.1/depiction');
	foreach my $photo (@photos)
		{ push @rv, ['PHOTO;VALUE=uri',$photo]; }

	# Logo
	my @logos = allOf($S, 'urn:ietf:rfc:2426#logo',
		'http://www.w3.org/2006/vcard/ns#logo',
		'http://www.w3.org/2006/03/hcard#logo',
		'http://www.w3.org/2001/vcard-rdf/3.0#LOGO',
		'http://xmlns.com/foaf/0.1/logo');
	foreach my $logo (@logos)
		{ push @rv, ['LOGO;VALUE=uri',$logo]; }

	# Sound
	my @sounds = allOf($S, 'urn:ietf:rfc:2426#sound',
		'http://www.w3.org/2006/vcard/ns#sound',
		'http://www.w3.org/2006/03/hcard#sound',
		'http://www.w3.org/2001/vcard-rdf/3.0#SOUND');
	foreach my $sound (@sounds)
		{ push @rv, ['SOUND;VALUE=uri',$sound]; }

	# Key (crypto)
	my @keys = allOf($S, 'urn:ietf:rfc:2426#key',
		'http://www.w3.org/2006/vcard/ns#key',
		'http://www.w3.org/2006/03/hcard#key',
		'http://www.w3.org/2001/vcard-rdf/3.0#KEY',
		'http://xmlns.com/wot/0.1/pubkeyAddress');
	foreach my $key (@keys)
		{ push @rv, ['KEY;VALUE=uri',$key]; }
	
	# Keys (WOT RDF schema)
	my @keys = allOfObj($S, 'http://xmlns.com/wot/0.1/hasKey');
	foreach my $key (@keys)
	{
		my $keySubject = findSubject($triples, $key->toString);
		if ($keySubject)
		{
			my $uri = firstOf($keySubject, 'http://xmlns.com/wot/0.1/pubkeyAddress');
			push @rv, ['KEY;VALUE=uri', $uri]
				if ($uri);
		}
	}

	# Notes
	my @notes = allOf($S, 'urn:ietf:rfc:2426#note',
		'http://www.w3.org/2006/vcard/ns#note',
		'http://www.w3.org/2006/03/hcard#note',
		'http://www.w3.org/2001/vcard-rdf/3.0#NOTE',
		'http://www.w3.org/2000/01/rdf-schema#comment',
		'http://purl.org/dc/terms/abstract',
		'http://purl.org/dc/terms/description',
		'http://purl.org/dc/elements/1.1/description',
		'http://xmlns.com/foaf/0.1/plan',
		'http://xmlns.com/foaf/0.1/geekcode');
	foreach my $note (@notes)
		{ push @rv, ['NOTE',$note]; }

	# Agents
	my @agents = allOfObj($S, 'urn:ietf:rfc:2426#agent');
	foreach my $agent (@agents)
	{
		my $agentSubject = findSubject($triples, $agent->toString);
		if ($agentSubject)
		{
			if (is($agentSubject))
			{
				my $a = outVCF($triples, $agentSubject->uri);
				push @rv, ['AGENT',$a];
			}
			else
			{
				my $a = firstOf($agentSubject,
					'urn:ietf:rfc:2426#fn',
					'http://www.w3.org/2006/vcard/ns#fn',
					'http://www.w3.org/2006/03/hcard#fn',
					'http://www.w3.org/2001/vcard-rdf/3.0#FN',
					'http://xmlns.com/foaf/0.1/fullName',
					'http://xmlns.com/foaf/0.1/name',
					'http://www.w3.org/2000/10/swap/pim/contact#knownAs',
					'http://purl.org/dc/terms/title',
					'http://purl.org/dc/elements/1.1/title');
				push @rv, ['AGENT;VALUE=text',$a] if ($a);
			}
		}
		elsif ($agent->type)
		{
			push @rv, ['AGENT;VALUE=uri',$agent->toString];
		}
		else
		{
			my $a = $agent->toString;
			$a =~ s/(^\s+|\s+$)//g;
			$a =~ s/\s+/ /g;
			push @rv, ['AGENT;VALUE=text',$a] if ($a);
		}
	}

	# TZ (timezone)
	my $tz = firstOf($S, 'urn:ietf:rfc:2426#tz',
		'http://www.w3.org/2006/vcard/ns#tz',
		'http://www.w3.org/2006/03/hcard#tz',
		'http://www.w3.org/2001/vcard-rdf/3.0#TZ');
	push @rv, ['TZ',$tz]
		if ($tz);

	# Bday (date of birth)
	my $bday = firstOf($S, 'urn:ietf:rfc:2426#bday',
		'http://www.w3.org/2006/vcard/ns#bday',
		'http://www.w3.org/2006/03/hcard#bday',
		'http://www.w3.org/2001/vcard-rdf/3.0#BDAY',
		'http://www.w3.org/2000/10/swap/pim/contact#birthday',
		'http://xmlns.com/foaf/0.1/birthday');
	push @rv, ['BDAY',$bday]
		if ($bday);
		
	# Dday (date of death)
	my $dday = firstOf($S, 'urn:ietf:rfc:2426#dday');
	push @rv, ['DDAY',$dday]
		if ($dday);
		
	# Gender
	my $gender = firstOf($S, 'urn:ietf:rfc:2426#gender',
		'http://xmlns.com/foaf/0.1/gender');
	if ($gender)
	{
		push @rv, ['GENDER',$gender];
	}
	else
	{
		if (inProp($S, 'http://www.w3.org/2000/10/swap/pim/contact#Male',
			'http://www.w3.org/1999/02/22-rdf-syntax-ns#type'))
		{
			push @rv, ['GENDER','M'];
		}

		if (inProp($S, 'http://www.w3.org/2000/10/swap/pim/contact#Female',
			'http://www.w3.org/1999/02/22-rdf-syntax-ns#type'))
		{
			push @rv, ['GENDER','F'];
		}
	}
	
	# Categories
	my @cats = allOfObj($S, 'urn:ietf:rfc:2426#category',
		'http://www.w3.org/2006/vcard/ns#category',
		'http://www.w3.org/2006/03/hcard#category',
		'http://www.w3.org/2001/vcard-rdf/3.0#CATEGORIES');
	my $catstring = '';
	foreach my $cat (@cats)
	{
		my $catSubject = findSubject($triples, $cat->toString);
		if ($catSubject)
		{
			my $category = firstOf($catSubject,
				'http://www.w3.org/1999/02/22-rdf-syntax-ns#value',
				'http://purl.org/dc/terms/description',
				'http://purl.org/dc/elements/1.1/description');
			$catstring .= ','.uc($category)
				if (length $category);
		}
		else
		{
			my $category = $cat->toString;
			$category =~ s/(^\s+|\s+$)//g;
			$category =~ s/\s+/ /g;
			$catstring .= ','.uc($category);
		}
	}
	$catstring =~ s/^\,//;
	push @rv, ['CATEGORIES',$catstring,1]
		if ($catstring);

	# Class
	my $class = uc(firstOf($S, 'urn:ietf:rfc:2426#class',
		'http://www.w3.org/2006/vcard/ns#class',
		'http://www.w3.org/2006/03/hcard#class',
		'http://www.w3.org/2001/vcard-rdf/3.0#CLASS'));
	$class =~ s/\s//g;
	push @rv, ['CLASS',$class]
		if ($class =~ /^(PUBLIC|PRIVATE|CONFIDENTIAL|X-)/);
		
	# Rev
	my $rev = uc(firstOf($S, 'urn:ietf:rfc:2426#rev',
		'http://www.w3.org/2006/vcard/ns#rev',
		'http://www.w3.org/2006/03/hcard#rev',
		'http://www.w3.org/2001/vcard-rdf/3.0#REV'));
	push @rv, ['REV',$rev]
		if ($rev);
		
	# Kind
	my $kind = firstOf($S, 'urn:ietf:rfc:2426#kind');
	push @rv, ['KIND',$kind]
		if ($kind);
		
	# Lang
	my $lang = firstOf($S, 'urn:ietf:rfc:2426#lang');
	push @rv, ['LANG',$lang]
		if ($lang);
		
	# Birth, place thereof
	my $birth = firstOf($S, 'urn:ietf:rfc:2426#birth');
	push @rv, ['BIRTH',$birth]
		if ($birth);
		
	# Death, place thereof
	my $death = firstOf($S, 'urn:ietf:rfc:2426#death');
	push @rv, ['DEATH',$death]
		if ($death);
		
	# Done.
	push @rv, ['END','VCARD'];
	
	# Convert to text/directory syntax.
	my $rv = '';
	foreach my $line (@rv)
	{
		my ($k, $v, $soft) = @{$line};
		
		if ($k =~ /VALUE.uri$/i && $v =~ /^data\:/)
		{
			my $uri = URI->new($v);
			$v = encode_base64($uri->data, '');
			$k =~ s/VALUE.uri$/ENCODING\=b/i;
			$k .= ';TYPE='.$uri->media_type
				if ($uri->media_type);
		}
		
		$v =~ s/\\/\\\\/g;
		$v =~ s/\r?\n/\\n/g;
		unless ($soft)
		{
			$v =~ s/\;/\\\;/g;
			$v =~ s/\,/\\\,/g;
		}
		
		$rv .= "$k:$v\r\n";
	}
	return $rv;
}


sub outName
{
	my $triples = shift;
	my $subject = shift;
	my $S = $triples->{$subject};
	
	my $fn = firstOf($S, 'urn:ietf:rfc:2426#fn',
		'http://www.w3.org/2006/vcard/ns#fn',
		'http://www.w3.org/2006/03/hcard#fn',
		'http://www.w3.org/2001/vcard-rdf/3.0#FN',
		'http://xmlns.com/foaf/0.1/fullName',
		'http://xmlns.com/foaf/0.1/name',
		'http://www.w3.org/2000/10/swap/pim/contact#knownAs',
		'http://purl.org/dc/terms/title',
		'http://purl.org/dc/elements/1.1/title');	
	if (length $fn)
		{ return $fn; }

	# Name components
	my $honorific_pfx = firstOf($S, 'urn:ietf:rfc:2426#honorific-prefix',
		'http://www.w3.org/2006/vcard/ns#honorific-prefix',
		'http://www.w3.org/2006/03/hcard#honofific-prefix',
		'http://www.w3.org/2001/vcard-rdf/3.0#Prefix',
		'http://xmlns.com/foaf/0.1/title',
		'http://www.w3.org/2000/10/swap/pim/contact#personalTitle');
	my $firstname = firstOf($S, 'urn:ietf:rfc:2426#given-name',
		'http://www.w3.org/2006/vcard/ns#given-name',
		'http://www.w3.org/2006/03/hcard#given-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#Given',
		'http://xmlns.com/foaf/0.1/givenname',
		'http://www.w3.org/2000/10/swap/pim/contact#firstName');
	my $surname = firstOf($S, 'urn:ietf:rfc:2426#family-name',
		'http://www.w3.org/2006/vcard/ns#family-name',
		'http://www.w3.org/2006/03/hcard#family-name',
		'http://www.w3.org/2001/vcard-rdf/3.0#Family',
		'http://xmlns.com/foaf/0.1/family_name',
		'http://xmlns.com/foaf/0.1/surname',
		'http://www.w3.org/2000/10/swap/pim/contact#lastName');
	my $honorific_sfx = firstOf($S, 'urn:ietf:rfc:2426#honorific-suffix',
		'http://www.w3.org/2006/vcard/ns#honorific-suffix',
		'http://www.w3.org/2006/03/hcard#honorific-suffix',
		'http://www.w3.org/2001/vcard-rdf/3.0#Suffix',
		'http://www.w3.org/2000/10/swap/pim/contact#personalSuffix');

	$fn = "$honorifix_pfx $firstname $surname $honorific_sfx";
	$fn =~ s/(^\s+|\s+$)//g;
	$fn =~ s/\s+/ /g;
	if (length $fn)
		{ return $fn; }
		
	# Nickname
	return firstOf($S, 'urn:ietf:rfc:2426#nickname',
		'http://www.w3.org/2006/vcard/ns#nickname',
		'http://www.w3.org/2006/03/hcard#nickname',
		'http://www.w3.org/2001/vcard-rdf/3.0#NICKNAME',
		'http://xmlns.com/foaf/0.1/nick',
		'http://www.w3.org/2000/10/swap/pim/contact#knownAs');
}

# Implements the atom "Person" construct, though actually we may be an
# organisation or other type of contact.
sub _AtomPerson
{
	my $triples = shift;
	my $subject = shift;
	my $atomtag = shift;
	my $indent;
	my $S = $triples->{$subject};

	$atomtag =~ /^(\s*)<(.*)>$/;
	$indent  = $1;
	$atomtag = $2;

	my $name  = outName($triples, $subject);
	my $email = firstOf($S,
		'urn:ietf:rfc:2426#email',
		'http://www.w3.org/2006/vcard/ns#email',
		'http://www.w3.org/2006/vcard/ns#workEmail',
		'http://www.w3.org/2006/vcard/ns#mobileEmail',
		'http://www.w3.org/2006/vcard/ns#unlabelledEmail',
		'http://www.w3.org/2006/03/hcard#email',
		'http://www.w3.org/2001/vcard-rdf/3.0#EMAIL',
		'http://www.w3.org/2000/10/swap/pim/contact#emailAddress',
		'http://xmlns.com/foaf/0.1/mbox',
		'http://xmlns.com/foaf/0.1/mbox_sha1sum');
	my $url   = firstOf($S, 'urn:ietf:rfc:2426#url',
		'http://www.w3.org/2006/vcard/ns#url',
		'http://www.w3.org/2006/03/hcard#url',
		'http://xmlns.com/foaf/0.1/homepage',
		'http://xmlns.com/foaf/0.1/weblog',
		'http://xmlns.com/foaf/0.1/workplaceHomepage',
		'http://xmlns.com/foaf/0.1/workInfoHomepage',
		'http://xmlns.com/foaf/0.1/schoolHomepage',
		'http://xmlns.com/foaf/0.1/tipjar',
		'http://www.w3.org/2000/10/swap/pim/contact#homepage',
		'http://www.w3.org/2000/10/swap/pim/contact#publicHomepage');
	
	$email =~ s/(^mailto:|\?.*$)//g;
	
	my $rv = "$indent<$atomtag\n$indent    xmlns:hcard=\"urn:ietf:rfc:2426#\"\n$indent    xmlns:foaf=\"http://xmlns.com/foaf/0.1/\">\n";
	$rv .=   "$indent\t<name>".encode_entities_numeric($name)."</name>\n" if ($name);
	$rv .=   "$indent\t<email>".encode_entities_numeric($email)."</email>\n" if ($email);
	$rv .=   "$indent\t<uri>".encode_entities_numeric($url)."</uri>\n" if ($url);
	foreach my $p ($S->properties)
	{
		next if ($p eq 'urn:ietf:rfc:2426#n'   # redundant
			|| $p eq 'urn:ietf:rfc:2426#fn'     # redundant
			|| $p eq 'urn:ietf:rfc:2426#url'    # redundant
			|| $p eq 'urn:ietf:rfc:2426#email'  # redundant
			|| $p eq 'urn:ietf:rfc:2426#adr'    # ugly
			|| $p eq 'urn:ietf:rfc:2426#geo'    # ugly
			|| $p eq 'urn:ietf:rfc:2426#agent'  # ugly
			|| $p eq 'urn:ietf:rfc:2426#org'    # ugly
			|| $p eq 'urn:ietf:rfc:2426#tel'    # ugly
			);
			
		my $P = $S->property($p);
		if ($P->namespace eq 'urn:ietf:rfc:2426#'
		||  $P->namespace eq 'http://xmlns.com/foaf/0.1/')
		{
			my $ns = 'hcard';
			$ns = 'foaf' if ($P->namespace eq 'http://xmlns.com/foaf/0.1/');
			
			foreach my $V (@{ $P->valObjects })
			{
				my $v = $V->toString;
				$rv .= "$indent\t<$ns:".$P->localname.">".encode_entities_numeric($v)
					."</$ns:".$P->localname.">\n";
			}
		}
	}
	$rv .= "$indent</$atomtag>\n";
	return $rv;
}

1;
