#!/usr/bin/perl

######################################################################
package Cognition::uF::figure;
######################################################################

use XML::LibXML qw(:all);
use strict;
use Cognition::MagicString;

BEGIN
{
	our $VERSION = 0.1;
	our @ISA = qw(Cognition::uF);
}


sub parse_all
{
	my $page   = shift;
	my $within = shift || $page->{DOM};
	my @nodes = Cognition::HTMLParser::Utils::searchClass('figure', $within);
	my @rv;
	
	foreach my $figure (@nodes)
	{
		my $figure_parsed = parse($page, $figure);
		push @rv, $figure_parsed;
	}
	
	return @rv;
} #/sub parse_all


sub uri
{
	my $this = shift;
	my $all  = shift;
	my @rv;

	if (length $this->{image})
	{
		push @rv, $this->{image};
	}
	if (length $this->{_id})
	{
		push @rv, $this->{_page}->abs_url.'#'.$this->{_id};
	}
	if (length $this->{imgid})
	{
		push @rv, $this->{_page}->abs_url.'#'.$this->{imgid};
	}
	push @rv, Cognition::HTMLParser::Utils::BNodeURI($this->{_dom}, 'Figure')
		unless @rv;
	# push @rv, Cognition::HTMLParser::Utils::XPointer($this->{_dom}, $this->{_page}->abs_url);
	
	return $rv[0] unless (defined $all);
	while ($all) { shift @rv; $all--; }
	return @rv;
}


sub metadata_ns
{
	my $this = shift;
	return { 
		'foaf'     => { nsurl=>'http://xmlns.com/foaf/0.1/',title=>'FOAF' },
		'dcterms'  => { nsurl=>'http://purl.org/dc/terms/',title=>'Dublin Core Terms' },
		'skos'     => { nsurl=>'http://www.w3.org/2004/02/skos/core#',title=>'Simple Knowledge Organization System' }
		};
} #/sub namespace


sub rdf_subject
{
	my $this    = shift;

	# Create RDF subject
	my $subject = Cognition::RDFModel::Subject->new($this->uri);
	$subject->addObject($this);
	$subject->setDom($this->{_dom});
	
	$subject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
		Cognition::RDFModel::Value->new('http://xmlns.com/foaf/0.1/Image', 1));

	if (length $this->{legend})
	{
		my $val = Cognition::RDFModel::Value->new($this->{legend});
		$subject->addVal('http://purl.org/dc/terms/description', $val);
	}
	
	if (length $this->{title})
	{
		my $val = Cognition::RDFModel::Value->new($this->{title});
		$subject->addVal('http://purl.org/dc/terms/title', $val);
	}
	
	if (length $this->{alt})
	{
		my $val = Cognition::RDFModel::Value->new($this->{alt});
		$subject->addVal('http://www.w3.org/2004/02/skos/core#altLabel', $val);
	}

	foreach my $V (@{ $this->{credit} })
	{
		my $val;
		if (ref $V && UNIVERSAL::can($V, 'uri')) 
			{ $val = Cognition::RDFModel::Value->new($V->uri, 1); }
		else
			{ $val = Cognition::RDFModel::Value->new($V); }
		
		$subject->addVal('http://purl.org/dc/terms/contributor', $val);
	}
	
	foreach my $V (@{ $this->{subject} })
	{
		my $val;
		if (ref $V && UNIVERSAL::can($V, 'uri')) 
			{ $val = Cognition::RDFModel::Value->new($V->uri, 1); }
		else
			{ $val = Cognition::RDFModel::Value->new($V); }
		
		$subject->addVal('http://xmlns.com/foaf/0.1/depicts', $val);
	}

	return $subject;
	
} #/sub rdf_subject


sub parse
{
	my $page = shift;
	my $rv = { '_dom'=>shift };
	
	my $figure_node = $rv->{'_dom'}->cloneNode(1);

	my $id = $figure_node->getAttribute('id');
	$rv->{_id} = $id if (length $id);
	$rv->{_page} = $page;
	
	# Expand microformat patterns.
	my $incl_iterations = 0;
	my $replacements = 1;
	while (($incl_iterations < Cognition::uF::MAX_INCLUDE_ITERATIONS) && $replacements)
	{
		$replacements = Cognition::uF::ufp_include($figure_node, $page) 
		              + Cognition::uF::ufp_include2($figure_node, $page);
		$incl_iterations++;
	}
	Cognition::uF::ufp_abbr($figure_node, $page);
	Cognition::uF::ufp_table($figure_node, $page);

	# Extract embedded hCards, which may be used as credits
	my @nested = Cognition::HTMLParser::Utils::searchClass('vcard', $figure_node);
	foreach my $h (@nested)
	{
		if ($h->getAttribute('class') =~ / (^|\s) credit (\s|$) /x)
			{ push @{ $rv->{credit} },  Cognition::uF::hCard::parse($page, $h); }	
		if ($h->getAttribute('class') =~ / (^|\s) subject (\s|$) /x)
			{ push @{ $rv->{subject} }, Cognition::uF::hCard::parse($page, $h); }
			
		my $newClass = $h->getAttribute('class');
		$newClass =~ s/\b(credit|subject)\b//gix;
		$h->setAttribute('class', $newClass);
	}

	# Now that we have reached here, no other composite microformats have
	# any business being nested within this root element. So in the interest
	# of Microformat opacity, let's destroy them.
	Cognition::uF::destroyer($figure_node);
	
	# Additional credits	
	my @value_nodes = Cognition::HTMLParser::Utils::searchClass('credit', $figure_node);
	foreach my $value_node (@value_nodes)
	{
		my $value = Cognition::HTMLParser::Utils::STRINGIFY($value_node, 'value');
		push @{ $rv->{'credit'} }, $value;
	}

	# Additional subjects	
	my @value_nodes = Cognition::HTMLParser::Utils::searchClass('subject', $figure_node);
	foreach my $value_node (@value_nodes)
	{
		my $value = Cognition::HTMLParser::Utils::STRINGIFY($value_node, 'value');
		push @{ $rv->{'subject'} }, $value;
	}

	# Categories
	my @tags = Cognition::uF::RelTag::parse_all($page, $figure_node);
	$rv->{category} = \@tags unless (!@tags);	
	
	# Find image
	my $img;
	if ($figure_node->localname eq 'img')
		{ $img = $figure_node; }
	else
	{
		my @images = Cognition::HTMLParser::Utils::searchClass('image', $figure_node);
		@images = $figure_node->getElementsByTagName('img')
			unless (@images);
		$img = $images[0];
	}
	
	# Legend
	my @value_nodes = Cognition::HTMLParser::Utils::searchClass('legend', $figure_node);
	my $value_node  = $value_nodes[0];
	if ($value_node)
	{
		my $value = Cognition::HTMLParser::Utils::STRINGIFY($value_node, 'value');
		$value = ms($img->getAttribute('title'), $img)
			if ($value_node==$img);
		$rv->{'legend'} = $value;
	}
	
	if ($img)
	{
		$rv->{image} =   ($page->abs_url($img->getAttribute('src')));
		$rv->{title} = ms($img->getAttribute('title'), $img);
		$rv->{imgid} =   ($img->getAttribute('id'));
		$rv->{alt}   = ms($img->getAttribute('alt'), $img);
	}
	
	bless $rv;	
	return $rv;
	
} #/sub parse


1; #/package Cognition::uF::figure




